with Ada.Unchecked_Conversion;
with Ada.Unchecked_Deallocation;

package body Parallel_Algorithms is
    
    -- ---------------------------------------------------------------
    -- Constants
    -- ---------------------------------------------------------------

    NUM_WORKERS: constant Positive := 4;
    DEFAULT_DELAY: constant Duration := 0.05;

    -- ---------------------------------------------------------------
    -- Types
    -- ---------------------------------------------------------------

    task type Worker_Type is 
        entry Start(Input: Array_Access_Type; 
                    Input_From: Natural;
                    Input_To: Natural);
    end Worker_Type;

    type Worker_Access_Type is access all Worker_Type;
    type Workers_Array_Type is array(1 .. NUM_WORKERS) of Worker_Access_Type;

    protected type Logger_Type is

        procedure Add(Partial_Sum: Item_Type);
        function  Get_Sum return Item_Type;
        function  Has_Finished return Boolean;
        procedure Reset;
        procedure Set_Worker_Finished;

    private

        Is_First: Boolean := True;
        Sum: Item_Type;
        Num_Workers_Running: Natural := NUM_WORKERS;

    end Logger_Type;
    
    -- ---------------------------------------------------------------
    -- Members
    -- ---------------------------------------------------------------

    Workers: Workers_Array_Type;
    Logger: Logger_Type;

    -- ---------------------------------------------------------------

    procedure Free is new Ada.Unchecked_Deallocation(
        Worker_Type, Worker_Access_Type);

    function Integer_To_Item_Type is new Ada.Unchecked_Conversion(
        Source => Integer,
        Target => Item_Type
    );
    
    -- ---------------------------------------------------------------
    -- Worker
    -- ---------------------------------------------------------------

    task body Worker_Type is
        Sum: Item_Type;
        A: Array_Access_Type;
        From: Natural;
        To: Natural;
    begin
        loop 
            accept Start(Input: Array_Access_Type; 
                         Input_From: Natural;
                         Input_To: Natural) do
                A := Input;
                From := Input_From;
                To := Input_To;
            end Start;

            if To >= From then
                Sum := A.all(From);
            end if;

            for I in From+1 .. To loop
                Sum := Sum + A.all(I);
            end loop;

            Logger.Add(Sum);
            Logger.Set_Worker_Finished;
        end loop;
    end Worker_Type;

    -- ---------------------------------------------------------------
    -- Logger
    -- ---------------------------------------------------------------
    
    protected body Logger_Type is

        procedure Add(Partial_Sum: Item_Type) is
        begin
            if Is_First then
                Sum := Partial_Sum;
                Is_First := False;
            else
                Sum := Sum + Partial_Sum;
            end if;
        end Add;

        -- ---------------------------------------------------------------

        function Get_Sum return Item_Type is
        begin
            return Sum;
        end Get_Sum;

        -- ---------------------------------------------------------------

        function Has_Finished return Boolean is
        begin
            return Num_Workers_Running = 0;
        end Has_Finished;

        -- ---------------------------------------------------------------

        procedure Reset is
        begin
            Is_First := True;
            Num_Workers_Running := NUM_WORKERS;
        end Reset;

        -- ---------------------------------------------------------------

        procedure Set_Worker_Finished is
        begin
            Num_Workers_Running := Num_Workers_Running - 1;
        end Set_Worker_Finished;

    end Logger_Type;
    
    -- ---------------------------------------------------------------
    -- Worker Management
    -- ---------------------------------------------------------------

    procedure Start_Workers(Input: Array_Access_Type) is
        From: Natural := Input.all'First;
        To:   Natural;
        Part_Length: constant Natural := Input.all'Length / NUM_WORKERS;
    begin
        for I in Workers'range loop
            To := From + Part_Length - 1;

            if I = Workers'Last then
                To := Input.all'Last;
            end if;

            Workers(I) := new Worker_Type;
            Workers(I).Start(Input, From, To);

            From := To + 1;
        end loop;
    end Start_Workers;

    -- ---------------------------------------------------------------

    procedure Destroy_Workers is
    begin
        for I in Workers'Range loop
            abort Workers(I).all;
            Free(Workers(I));
        end loop;
    end Destroy_Workers;

    -- ---------------------------------------------------------------

    function Is_Empty(Input: Array_Access_Type) return Boolean is
    begin
        return (Input = null) or else (Input.all'Length = 0);
    end Is_Empty;

    -- ---------------------------------------------------------------

    function Do_Sequential_Sum(Input: Array_Access_Type) return Item_Type is
        Sum: Item_Type := Input.all(Input.all'First);
    begin
        for I in Input.all'First+1 .. Input.all'Last loop
            Sum := Sum + Input.all(I);
        end loop;

        return Sum;
    end Do_Sequential_Sum;

    -- ---------------------------------------------------------------

    function Do_Parallel_Sum(Input: Array_Access_Type) return Item_Type is
    begin
        Logger.Reset;
        Start_Workers(Input);

        while not Logger.Has_Finished loop
            delay DEFAULT_DELAY;
        end loop;

        Destroy_Workers;
        return Logger.Get_Sum;
    end Do_Parallel_Sum;

    -- ---------------------------------------------------------------
    -- Public API
    -- ---------------------------------------------------------------
    
    procedure Parallel_Sum(Input: Array_Access_Type;
                           Result: out Item_Type) is
    begin
        if Is_Empty(Input) then
            Result := Integer_To_Item_Type(0);
        elsif Input.all'Length < NUM_WORKERS then
            Result := Do_Sequential_Sum(Input);
        else
            Result := Do_Parallel_Sum(Input);
        end if;
    end Parallel_Sum;

end Parallel_Algorithms;