with Ada.Task_Identification;
with Ada.Text_IO;
with Ada.Unchecked_Deallocation;
with GNAT.Command_Line;
with GNAT.Strings;

with Array_IO;
with Parallel_Algorithms;

use Ada.Task_Identification;
use Ada.Text_IO;
use GNAT.Command_Line;
use GNAT.Strings;

procedure Main is
	
    -- -----------------------------------------------------
    -- Constants
    -- -----------------------------------------------------

    ENVIRONMENT_TASK: constant Task_ID := Current_Task;
    NUM_ELEMENTS_TO_LOG: constant Natural := 5;

    type Item_Type is new Integer;
    for Item_Type'Size use Integer'Size;
    
    package Parallel_Integer_Algorithms is new Parallel_Algorithms(
        Item_Type => Item_Type, 
        "<" => "<",
        "=" => "="
    );
    use Parallel_Integer_Algorithms;
    
    package Integer_IO is new Array_IO(
        Item_Type => Item_Type, 
        Array_Type => Array_Type, 
        Array_Access_Type => Array_Access_Type
    );
    use Integer_IO;
    
    procedure Free is new Ada.Unchecked_Deallocation(
        Array_Type, Array_Access_Type);
    
    
    -- -----------------------------------------------------
    -- Members
    -- -----------------------------------------------------

    type Context_Type is record
        In_File_Path: aliased String_Access;
        Out_File_Path: aliased String_Access;
        Max_Run_Time: aliased Integer;
    end record;

    Config: Command_Line_Configuration;
    Context: Context_Type;

    Input_Array: Array_Access_Type;
    Result_Array: Array_Access_Type;
    
    -- -----------------------------------------------------
    -- Helper functions
    -- -----------------------------------------------------
    
	procedure Config_Command_Line(Config  : in out Command_Line_Configuration;
                                  Context : in out Context_Type) is
	begin
        Define_Switch(
            Config, 
            Context.In_File_Path'access, 
            "-i:",
            Help => "File path of an input file containing one integer per line."
        );
        Define_Switch(
            Config, 
            Context.Out_File_Path'access, 
            "-o:",
            Help => "File path of desired output file."
        );
        Define_Switch(
            Config, 
            Context.Max_Run_Time'access, 
            "-t:",
            Initial => 5,
            Default => 5,
            Help => "Maximum run time of the program (seconds). Default: 5."
        );
        Getopt(Config);
	end Config_Command_Line;

    -- ---------------------------------------------------------------

    procedure Print_Usage is
    begin
        Put_Line("Sorts an array of integers from a given file.");
        Put_Line(
            "Usage: ./main -i <in_file_path> -o <out_file_path> -t <max_time>"
        );
        Put_Line("-<in_file_path>: Relative path to a text file with one");
        Put_Line(" integer per line.");
        Put_Line("-<out_file_path>: Relative path to the result file.");
        Put_Line("-<max_time>: Max. run time of the program.");
        Put_Line("Pressing q or Q stops the program immediately.");
    end Print_Usage;

    -- ---------------------------------------------------------------

    procedure Log_Array(A: Array_Access_Type; 
                        Max_Num_Elements_To_Log: Natural) is 
        Num_Elements_To_Log: Natural := Max_Num_Elements_To_Log;
    begin
        if A = null then
            Put_Line("Array length: 0");
            return;
        end if;

        Put_Line("Array length: " & A.all'Length'Img);

        if A.all'Length < Max_Num_Elements_To_Log then
            Num_Elements_To_Log := A.all'Length;
        end if;

        for I in A.all'First .. A.all'First + Num_Elements_To_Log loop
            Put_Line(I'Img & ": " & A.all(I)'Img);
        end loop;

        Put_Line("...");

        for I in A.all'Last - Num_Elements_To_Log + 1 .. A.all'Last loop
            Put_Line(I'Img & ": " & A.all(I)'Img);
        end loop;
    end Log_Array;

    -- ---------------------------------------------------------------

    procedure Read_Array(Input_Array:  in out Array_Access_Type; 
                         Result_Array: in out Array_Access_Type; 
                         Path: String) is
    begin
        Input_Array  := Read_File(Path);
        Result_Array := new Array_Type(Input_Array'range);
    end Read_Array;

    -- ---------------------------------------------------------------

    procedure Write_Array(A: Array_Access_Type; Path: String) is
    begin
        Write_To_File(Path, A.all);
    end Write_Array;

    -- ----------------------------------------------------
    -- Tasks
    -- ----------------------------------------------------

    task User_Input_Task;
    task body User_Input_Task is
        Available: Boolean;
        Char: Character;
    begin
        loop
            Get_Immediate(Char, Available);

            if Available and then 
                (Char = 'q' or else Char = 'Q') then
                Abort_Task(ENVIRONMENT_TASK);
            else
                delay 0.1;
            end if;
        end loop;
    end User_Input_Task;

begin
    Config_Command_Line(Config, Context);

    select
        delay Duration(Context.Max_Run_Time);
    then abort
        Read_Array(Input_Array, Result_Array, Context.In_File_Path.all);
        Log_Array(Input_Array, NUM_ELEMENTS_TO_LOG);
        Parallel_Merge_Sort(Input_Array, Result_Array);
        Log_Array(Result_Array, NUM_ELEMENTS_TO_LOG);
        Write_Array(Result_Array, Context.Out_File_Path.all);

        Free(Input_Array);
        Free(Result_Array);
    end select;

    Abort_Task(ENVIRONMENT_TASK);
exception
    when GNAT.Command_Line.Exit_From_Command_Line => Print_Usage;
    when others => Print_Usage;
end Main;
