with Ada.Text_IO;
with Ada.Unchecked_Deallocation;

use Ada.Text_IO;

package body Parallel_Algorithms is
    pragma Unreferenced("=");

    -- ---------------------------------------------------------------
    -- Constants
    -- ---------------------------------------------------------------

    NUM_WORKERS: constant Positive := 4;
    DEFAULT_DELAY: constant Duration := 0.05;

    -- ---------------------------------------------------------------
    -- Types
    -- ---------------------------------------------------------------

    task type Worker_Type is 
        entry Start(Input: Array_Access_Type; From_In: Natural; To_In: Natural);
    end Worker_Type;

    type Worker_Access_Type is access all Worker_Type;
    type Workers_Array_Type is array(1 .. NUM_WORKERS) of Worker_Access_Type;

    protected type Logger_Type is

        function  Has_Finished return Boolean;
        procedure Reset;
        procedure Set_Worker_Finished;

    private

        Is_First: Boolean := True;
        Num_Workers_Running: Natural := NUM_WORKERS;

    end Logger_Type;
    
    -- ---------------------------------------------------------------
    -- Members
    -- ---------------------------------------------------------------

    Workers: Workers_Array_Type;
    Logger: Logger_Type;

    -- ---------------------------------------------------------------

    procedure Free is new Ada.Unchecked_Deallocation(
        Worker_Type, Worker_Access_Type);

    procedure Free is new Ada.Unchecked_Deallocation(
        Array_Type, Array_Access_Type);

    -- ---------------------------------------------------------------
    -- Worker
    -- ---------------------------------------------------------------

    procedure Merge(Left: Array_Type; Right: Array_Type; 
                    Result: out Array_Type) is
        Left_Index: Natural := Left'First;
        Right_Index: Natural := Right'First;
        Result_Index: Natural := Result'First;
    begin
        while (Left_Index  <= Left'Last) and then 
              (Right_Index <= Right'Last) loop
            if Left(Left_Index) < Right(Right_Index) then
                Result(Result_Index) := Left(Left_Index);
                Left_Index := Left_Index + 1;
            else
                Result(Result_Index) := Right(Right_Index);
                Right_Index := Right_Index + 1;
            end if;
            Result_Index := Result_Index + 1;
        end loop;
        
        if Left_Index <= Left'Last then
            Result(Result_Index .. Result'Last) 
                := Left(Left_Index .. Left'Last);
        elsif Right_Index <= Right'Last then
            Result(Result_Index .. Result'Last) 
                := Right(Right_Index .. Right'Last);
        end if;
    end Merge;
    
    -- ---------------------------------------------------------------

    procedure Sort(A: in out Array_Type) is
    begin
        if A'Length <= 1 then
            return;
        end if;
        
        declare
            Middle: constant Natural := A'First + (A'Length / 2);
            Left : Array_Access_Type := new Array_Type(A'First .. Middle-1);
            Right: Array_Access_Type := new Array_Type(Middle  .. A'Last);
        begin
            Left(Left'range) := A(Left'range);
            Right(Right'range) := A(Right'range);
            
            Sort(Left.all);
            Sort(Right.all);
            Merge(Left.all, Right.all, A);

            Free(Left);
            Free(Right);
        end;
    end Sort;

    -- ---------------------------------------------------------------

    task body Worker_Type is
        A: Array_Access_Type;
        From: Natural;
        To: Natural;
    begin
        accept Start(Input: Array_Access_Type;
                     From_In: Natural;
                     To_In: Natural) do
            A := Input;
            From := From_In;
            To := To_In;
        end Start;

        Sort(A.all(From .. To));

        Logger.Set_Worker_Finished;
    end Worker_Type;

    -- ---------------------------------------------------------------
    -- Logger
    -- ---------------------------------------------------------------
    
    protected body Logger_Type is

        -- ---------------------------------------------------------------

        function Has_Finished return Boolean is
        begin
            return Num_Workers_Running = 0;
        end Has_Finished;

        -- ---------------------------------------------------------------

        procedure Reset is
        begin
            Is_First := True;
            Num_Workers_Running := NUM_WORKERS;
        end Reset;

        -- ---------------------------------------------------------------

        procedure Set_Worker_Finished is
        begin
            Num_Workers_Running := Num_Workers_Running - 1;
        end Set_Worker_Finished;

    end Logger_Type;
    
    -- ---------------------------------------------------------------
    -- Parallel sort workers management
    -- ---------------------------------------------------------------

    procedure Start_Workers(Input: Array_Access_Type) is
        From: Natural := Input.all'First;
        To:   Natural;
        Part_Length: constant Natural := Input.all'Length / NUM_WORKERS;
    begin
        for I in Workers'range loop
            To := From + Part_Length - 1;

            if I = Workers'Last then
                To := Input.all'Last;
            end if;

            Workers(I) := new Worker_Type;
            Workers(I).Start(Input, From, To);

            From := To + 1;
        end loop;
    end Start_Workers;

    -- ---------------------------------------------------------------

    procedure Merge_Workers(Input: Array_Access_Type) is
        Part_Length: constant Natural := Input.all'Length / NUM_WORKERS;
        A1: Array_Access_Type := new Array_Type(
                Input'First .. 
                Input'First + Part_Length - 1);
        A2: Array_Access_Type := new Array_Type(
            Input'First + Part_Length .. 
            Input'First + 2 * Part_Length - 1);
        A3: Array_Access_Type := new Array_Type(
            Input'First + 2 * Part_Length .. 
            Input'First + 3 * Part_Length - 1);
        A4: Array_Access_Type := new Array_Type(
            Input'First + 3 * Part_Length .. 
            Input'Last);
        A12: Array_Access_Type := new Array_Type(
            Input'First .. 
            Input'First + 2 * Part_Length - 1);
        A34: Array_Access_Type := new Array_Type(
            Input'First + 2 * Part_Length .. Input'Last);
    begin
        A1.all(A1'range) := Input.all(A1'range);
        A2.all(A2'range) := Input.all(A2'range);
        A3.all(A3'range) := Input.all(A3'range);
        A4.all(A4'range) := Input.all(A4'range);

        Merge(A1.all,  A2.all,  A12.all);
        Merge(A3.all,  A4.all,  A34.all);
        Merge(A12.all, A34.all, Input.all);

        Free(A1);
        Free(A2);
        Free(A3);
        Free(A4);
        Free(A12);
        Free(A34);
    end Merge_Workers;

    -- ---------------------------------------------------------------

    procedure Destroy_Workers is
    begin
        for I in Workers'Range loop
            abort Workers(I).all;
            Free(Workers(I));
        end loop;
    end Destroy_Workers;

    -- ---------------------------------------------------------------

    procedure Do_Parallel_Sort(Input: Array_Access_Type;
                               Result: Array_Access_Type) is
    begin
        Logger.Reset;

        Result.all(Result.all'range) := Input.all(Input.all'range);
        Put_Line("Starting parallel sort");
        Start_Workers(Result);

        while not Logger.Has_Finished loop
            delay DEFAULT_DELAY;
        end loop;

        Put_Line("Finished sort");
        Put_Line("Started merging");
        Merge_Workers(Result);
        Put_Line("Finished Merging");
        Destroy_Workers;
        Put_Line("Destroyed Workers");
    end Do_Parallel_Sort;

    -- ---------------------------------------------------------------
    -- Sequential sort
    -- ---------------------------------------------------------------

    function Is_Empty(Input: Array_Access_Type) return Boolean is
    begin
        return (Input = null) or else (Input.all'Length = 0);
    end Is_Empty;

    -- ---------------------------------------------------------------

    procedure Swap(Input: Array_Access_Type; I: Natural; J: Natural) is
        Tmp: constant Item_Type := Input.all(I);
    begin
        Input.all(I) := Input.all(J);
        Input.all(J) := Tmp;
    end Swap;

    -- ---------------------------------------------------------------

    function Find_Minimum_Index(Input: Array_Access_Type; From: Natural) 
        return Natural is 
        Minimum_Index: Natural := From;
    begin
        if From = Natural'Last or From = Input.all'Last then
            return Minimum_Index;
        end if;

        for I in From+1 .. Input.all'Last loop
            if Input.all(I) < Input.all(Minimum_Index) then
                Minimum_Index := I;
            end if;
        end loop;

        return Minimum_Index;
    end Find_Minimum_Index;

    -- ---------------------------------------------------------------

    procedure Do_Sequential_Sort(Input: Array_Access_Type; 
                                 Result: Array_Access_Type) is
        J: Natural;
    begin
        Result.all(Result.all'range) := Input.all(Input.all'range);

        for I in Result.all'range loop
            J := Find_Minimum_Index(Result, I);
            Swap(Result, I, J);
        end loop;
    end Do_Sequential_Sort;

    -- ---------------------------------------------------------------
    -- Public API
    -- ---------------------------------------------------------------
    
    procedure Parallel_Merge_Sort(Input: Array_Access_Type;
                                  Result: Array_Access_Type) is
    begin
        if Is_Empty(Input) then
            Result.all := Input.all;
        elsif Input.all'Length < NUM_WORKERS then
            Do_Sequential_Sort(Input, Result);
        else
            Do_Parallel_Sort(Input, Result);
        end if;
        Put_Line("Ended Parallel_Merge_Sort");
    end Parallel_Merge_Sort;

end Parallel_Algorithms;