with Ada.Unchecked_Deallocation;
with AUnit.Assertions;
with Parallel_Algorithms;

use AUnit.Assertions;

package body Parallel_Algorithms_Tests is

    package Integer_Algorithms is new Parallel_Algorithms(
        Item_Type  => Integer,
        "<"        => "<",
        "="        => "="
    );
    use Integer_Algorithms;

    procedure Free is new Ada.Unchecked_Deallocation(
        Array_Type, Array_Access_Type);

    Mixed_List: constant Array_Type := (23, 37, 1, 12, 25, 9, 6, 18, 42, 13);

    -- ---------------------------------------------------------------

    function Create_Array(Num_Items: Natural) return Array_Access_Type is
        A: constant Array_Access_Type := new Array_Type(0 .. Num_Items-1);
    begin
        return A;
    end Create_Array;

    -- ---------------------------------------------------------------

    procedure Do_Free(A: in out Array_Access_Type; 
                      B: in out Array_Access_Type;  
                      Success: in out Boolean) is
    begin
        Free(A);
        Free(B);
        Success := True;
    end Do_Free;

    -- ---------------------------------------------------------------

    procedure Fill_Array_Sorted(A: Array_Access_Type) is
    begin
        for I in A.all'range loop
            A.all(I) := I;
        end loop;
    end Fill_Array_Sorted;

    -- ---------------------------------------------------------------

    procedure Fill_Array_Mixed(A: Array_Access_Type) is
        J: Natural;
    begin
        for I in A'range loop
            J  := I mod Mixed_List'Length;
            A.all(I) := Mixed_List(J);
        end loop;
    end Fill_Array_Mixed;

    -- ---------------------------------------------------------------

    procedure Fill_Array_With_Equal_Items(A: Array_Access_Type) is
    begin
        for I in A'range loop
            A.all(I) := 42;
        end loop;
    end Fill_Array_With_Equal_Items;

    -- ---------------------------------------------------------------

    procedure Fill_Array_Reverse_Sorted(A: Array_Access_Type) is
    begin
        for I in A.all'range loop
            A.all(I) := A.all'Last - I + A.all'First;
        end loop;
    end Fill_Array_Reverse_Sorted;

    -- ---------------------------------------------------------------

    function Is_Empty(A: Array_Access_Type) return Boolean is
    begin
        return (A = null) or else (A.all'Length > 0);
    end Is_Empty;

    -- ---------------------------------------------------------------

    function Is_Sorted(A: Array_Access_Type) return Boolean is
    begin
        if Is_Empty(A) then
            return True;
        end if;

        for I in A.all'First .. A.all'Last-1 loop
            if A.all(I) > A.all(I+1) then
                return False;
            end if;
        end loop;

        return True;
    end Is_Sorted;

    -- ---------------------------------------------------------------

    function Contains(A: Array_Access_Type; Item: Natural) return Boolean is
    begin
        for I in A.all'range loop
            if A.all(I) = Item then
                return True;
            end if;
        end loop;

        return False;
    end Contains;

    -- ---------------------------------------------------------------

    function Are_Sets_Equal(A: Array_Access_Type; B: Array_Access_Type) 
        return Boolean is
    begin
        if (A = null and then B /= null) or
            (B = null and then A /= null) then
            return False;
        end if;

        if A.all'First /= B.all'First or else A.all'Last /= B.all'Last then
            return False;
        end if;

        for I in B.all'range loop
            if not Contains(A, B.all(I)) then
                return False;
            end if;
        end loop;

        return True;
    end Are_Sets_Equal;

    -- ---------------------------------------------------------------

    procedure Test_Array_Sorted(Num_Items: Natural) is
        A:     Array_Access_Type := Create_Array(Num_Items);
        A_Old: Array_Access_Type := Create_Array(Num_Items);
        Has_Freed: Boolean := False;
    begin
        Fill_Array_Sorted(A_Old);
        Parallel_Merge_Sort(A_Old, A);
        Assert(Is_Sorted(A), "Array was not sorted when filled sorted");
        Assert(Are_Sets_Equal(A, A_Old), 
            "Array lacks elements after sort when filled sorted");

        Do_Free(A, A_Old, Has_Freed);
    exception
        when others => 
            if not Has_Freed then 
                Do_Free(A, A_Old, Has_Freed);
            end if;
    end Test_Array_Sorted;

    -- ---------------------------------------------------------------

    procedure Test_Array_Mixed(Num_Items: Natural) is
        A:     Array_Access_Type := Create_Array(Num_Items);
        A_Old: Array_Access_Type := Create_Array(Num_Items);
        Has_Freed: Boolean := False;
    begin
        Fill_Array_Mixed(A_Old);
        Parallel_Merge_Sort(A_Old, A);

        Assert(Is_Sorted(A), "Array was not sorted when filled mixed");
        Assert(Are_Sets_Equal(A, A_Old), 
            "Array lacks elements after sort when filled mixed");
        
        Do_Free(A, A_Old, Has_Freed);
    exception
        when others => 
            if not Has_Freed then 
                Do_Free(A, A_Old, Has_Freed);
            end if;
    end Test_Array_Mixed;

    -- ---------------------------------------------------------------
    
    procedure Test_Array_With_Duplicates(Num_Items: Natural) is
        A:     Array_Access_Type := Create_Array(Num_Items);
        A_Old: Array_Access_Type := Create_Array(Num_Items);
        Has_Freed: Boolean := False;
    begin
        Fill_Array_With_Equal_Items(A_Old);
        Parallel_Merge_Sort(A_Old, A);
        Assert(Is_Sorted(A), 
            "Array was not sorted when filled with duplicates");
        Assert(Are_Sets_Equal(A, A_Old), "Array lacks elements "
            & "after sort when filled with duplicates");

        Do_Free(A, A_Old, Has_Freed);
    exception
        when others => 
            if not Has_Freed then 
                Do_Free(A, A_Old, Has_Freed);
            end if;
    end Test_Array_With_Duplicates;

    -- ---------------------------------------------------------------
    
    procedure Test_Array_Reverse_Sorted(Num_Items: Natural) is
        A:     Array_Access_Type := Create_Array(Num_Items);
        A_Old: Array_Access_Type := Create_Array(Num_Items);
        Has_Freed: Boolean := False;
    begin
        Fill_Array_Reverse_Sorted(A_Old);
        Parallel_Merge_Sort(A_Old, A);
        Assert(Is_Sorted(A), 
            "Array was not sorted when filled in reverse order");
        Assert(Are_Sets_Equal(A, A_Old), "Array lacks elements after "
            & "sort when filled in reverse order");

        Do_Free(A, A_Old, Has_Freed);
    exception
        when others => 
            if not Has_Freed then 
                Do_Free(A, A_Old, Has_Freed);
            end if;
    end Test_Array_Reverse_Sorted;

    -- ---------------------------------------------------------------

    procedure Test_Sort_Empty_Sorted(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_Sorted(0);
    end Test_Sort_Empty_Sorted;

    -- ---------------------------------------------------------------

    procedure Test_Sort_Empty_Mixed(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_Mixed(0);
    end Test_Sort_Empty_Mixed;

    -- ---------------------------------------------------------------

    procedure Test_Sort_Empty_With_Duplicates(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_With_Duplicates(0);
    end Test_Sort_Empty_With_Duplicates;

    -- ---------------------------------------------------------------

    procedure Test_Sort_Empty_Reverse_Sorted(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_Reverse_Sorted(0);
    end Test_Sort_Empty_Reverse_Sorted;

    -- ---------------------------------------------------------------
    
    procedure Test_Sort_Single_Item_Sorted(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_Sorted(1);
    end Test_Sort_Single_Item_Sorted;

    -- ---------------------------------------------------------------
    
    procedure Test_Sort_Single_Item_Mixed(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_Mixed(1);
    end Test_Sort_Single_Item_Mixed;

    -- ---------------------------------------------------------------
    
    procedure Test_Sort_Single_Item_With_Duplicates(T: in out Test_Case'Class) 
    is
        pragma Unreferenced(T);
    begin
        Test_Array_With_Duplicates(1);
    end Test_Sort_Single_Item_With_Duplicates;

    -- ---------------------------------------------------------------
    
    procedure Test_Sort_Single_Item_Reverse_Sorted(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_Reverse_Sorted(1);
    end Test_Sort_Single_Item_Reverse_Sorted;

    -- ---------------------------------------------------------------

    procedure Test_Sort_Few_Items_Sorted(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_Sorted(3);
    end Test_Sort_Few_Items_Sorted;

    -- ---------------------------------------------------------------

    procedure Test_Sort_Few_Items_Mixed(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_Mixed(3);
    end Test_Sort_Few_Items_Mixed;

    -- ---------------------------------------------------------------

    procedure Test_Sort_Few_Items_With_Duplicates(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_With_Duplicates(3);
    end Test_Sort_Few_Items_With_Duplicates;

    -- ---------------------------------------------------------------

    procedure Test_Sort_Few_Items_Reverse_Sorted(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_Reverse_Sorted(3);
    end Test_Sort_Few_Items_Reverse_Sorted;

    -- ---------------------------------------------------------------

    procedure Test_Sort_Some_Items_Sorted(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_Sorted(10);
    end Test_Sort_Some_Items_Sorted;

    -- ---------------------------------------------------------------

    procedure Test_Sort_Some_Items_Mixed(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_Mixed(10);
    end Test_Sort_Some_Items_Mixed;

    -- ---------------------------------------------------------------

    procedure Test_Sort_Some_Items_With_Duplicates(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_With_Duplicates(10);
    end Test_Sort_Some_Items_With_Duplicates;

    -- ---------------------------------------------------------------

    procedure Test_Sort_Some_Items_Reverse_Sorted(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_Reverse_Sorted(10);
    end Test_Sort_Some_Items_Reverse_Sorted;

    -- ---------------------------------------------------------------

    procedure Test_Sort_Many_Items_Sorted(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_Sorted(1_000);
    end Test_Sort_Many_Items_Sorted;

    -- ---------------------------------------------------------------

    procedure Test_Sort_Many_Items_Mixed(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_Mixed(1_000);
    end Test_Sort_Many_Items_Mixed;

    -- ---------------------------------------------------------------

    procedure Test_Sort_Many_Items_With_Duplicates(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_With_Duplicates(1_000);
    end Test_Sort_Many_Items_With_Duplicates;

    -- ---------------------------------------------------------------

    procedure Test_Sort_Many_Items_Reverse_Sorted(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Test_Array_Reverse_Sorted(1_000);
    end Test_Sort_Many_Items_Reverse_Sorted;

    -- ---------------------------------------------------------------

    procedure Test_Large_Array_Mixed(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Num_Items: constant Natural := 10_000_000;
        A:     Array_Access_Type := Create_Array(Num_Items);
        A_Old: Array_Access_Type := Create_Array(Num_Items);
        Has_Freed: Boolean := False;
    begin
        Fill_Array_Mixed(A_Old);
        Parallel_Merge_Sort(A_Old, A);

        Assert(Is_Sorted(A), "Array was not sorted when filled mixed");

        Do_Free(A, A_Old, Has_Freed);
    exception
        when others => 
            if not Has_Freed then 
                Do_Free(A, A_Old, Has_Freed);
            end if;
    end Test_Large_Array_Mixed;

    -- ---------------------------------------------------------------
    
    procedure Register_Tests(T: in out Test) is
        use AUnit.Test_Cases.Registration;
    begin
        Register_Routine(T, Test_Sort_Empty_Sorted'Access, 
            "Test_Sort_Empty_Sorted");
        Register_Routine(T, Test_Sort_Empty_Mixed'Access, 
            "Test_Sort_Empty_Mixed");
        Register_Routine(T, Test_Sort_Empty_With_Duplicates'Access, 
            "Test_Sort_Empty_With_Duplicates");
        Register_Routine(T, Test_Sort_Empty_Reverse_Sorted'Access, 
            "Test_Sort_Empty_Reverse_Sorted");
        Register_Routine(T, Test_Sort_Single_Item_Sorted'Access, 
            "Test_Sort_Single_Item_Sorted");
        Register_Routine(T, Test_Sort_Single_Item_Mixed'Access, 
            "Test_Sort_Single_Item_Mixed");
        Register_Routine(T, Test_Sort_Single_Item_With_Duplicates'Access, 
            "Test_Sort_Single_Item_With_Duplicates");
        Register_Routine(T, Test_Sort_Single_Item_Reverse_Sorted'Access, 
            "Test_Sort_Single_Item_Reverse_Sorted");
        Register_Routine(T, Test_Sort_Few_Items_Sorted'Access, 
            "Test_Sort_Few_Items_Sorted");
        Register_Routine(T, Test_Sort_Few_Items_Mixed'Access, 
            "Test_Sort_Few_Items_Mixed");
        Register_Routine(T, Test_Sort_Few_Items_With_Duplicates'Access, 
            "Test_Sort_Few_Items_With_Duplicates");
        Register_Routine(T, Test_Sort_Few_Items_Reverse_Sorted'Access, 
            "Test_Sort_Few_Items_Reverse_Sorted");
        Register_Routine(T, Test_Sort_Some_Items_Sorted'Access, 
            "Test_Sort_Some_Items_Sorted");
        Register_Routine(T, Test_Sort_Some_Items_Mixed'Access, 
            "Test_Sort_Some_Items_Mixed");
        Register_Routine(T, Test_Sort_Some_Items_With_Duplicates'Access, 
            "Test_Sort_Some_Items_With_Duplicates");
        Register_Routine(T, Test_Sort_Some_Items_Reverse_Sorted'Access, 
            "Test_Sort_Some_Items_Reverse_Sorted");
        Register_Routine(T, Test_Sort_Many_Items_Sorted'Access, 
            "Test_Sort_Many_Items_Sorted");
        Register_Routine(T, Test_Sort_Many_Items_Mixed'Access, 
            "Test_Sort_Many_Items_Mixed");
        Register_Routine(T, Test_Sort_Many_Items_With_Duplicates'Access, 
            "Test_Sort_Many_Items_With_Duplicates");
        Register_Routine(T, Test_Sort_Many_Items_Reverse_Sorted'Access, 
            "Test_Sort_Many_Items_Reverse_Sorted");
        Register_Routine(T, Test_Large_Array_Mixed'Access, 
            "Test_Large_Array_Mixed");
    end Register_Tests;
    
    -- ---------------------------------------------------------------

    function Name(T: Test) return Test_String is
        pragma Unreferenced(T);
    begin
        return Format("Parallel_Algorithms_Tests");
    end Name;

end Parallel_Algorithms_Tests;
